/*
    View64 C64 image viewer
    (c) 1995-2013 Soci/Singular (soci at c64.rulez.org)

    $Id: view64.c 60 2015-06-07 09:22:23Z soci $

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/

#include <stdio.h>
#include <allegro.h>

#include "libview64.h"
#include "tvrender.h"
#include "view64_config.h"
#include "main.h"
#include "cfgpath.h"
#include "ini.h"

static arguments_t arguments;
static int lastzoom;

BEGIN_DIGI_DRIVER_LIST
END_DIGI_DRIVER_LIST
BEGIN_MIDI_DRIVER_LIST
END_MIDI_DRIVER_LIST
BEGIN_JOYSTICK_DRIVER_LIST
END_JOYSTICK_DRIVER_LIST

typedef struct Rect {
    int x, y, w, h;
} Rect;

static Rect vidmode, windim, olddim, lastblit;
static int desktopdepth;

static BITMAP *create_surface(view64_screen *rscreen, unsigned int vmode, int scale, const rgb_color *palette, PALETTE *screen_palette, int width, int height) {
    BITMAP *pic;
    tvrender_image *out = create_tvpicture(rscreen, vmode, scale, palette);
    int use_palette, x, y;
    if (out == NULL) {
        fprintf(stderr, "Couldn't create picture\n");
        return NULL;
    }
    use_palette = out->palette && out->colors <= 256;
    x = rscreen->w * scale;
    y = rscreen->h * scale;
    pic = create_bitmap_ex(use_palette ? 8 : 24, x < width ? x : width, y < height ? y : height);
    if (pic == NULL) {
        free_tv_pixels(out);
        fprintf(stderr, "Couldn't create bitmap.\n");
        return NULL;
    }
    if (use_palette) {
        int y2 = y > height ? (y - height) / 2 : 0;
        int x2 = x > width ? (x - width) / 2 : 0;
        rgb_color *pal = out->palette;
        for (x = 0; x < out->colors; x++, pal++) {
            RGB *col = &(*screen_palette)[x];
            col->r = pal->r >> 2;
            col->g = pal->g >> 2;
            col->b = pal->b >> 2;
        }
        if (get_color_depth() <= 8 && rgb_map) {
            lastblit.x = lastblit.y = 0; lastblit.w = SCREEN_W; lastblit.h = SCREEN_H;
            rgb_map = NULL;
        }
        for (y = 0; y < pic->h; y++, y2++) {
            unsigned char *line = (unsigned char *)pic->line[y];
            unsigned char *e, *e2 = line + (scale - (x2 % scale)); e = line + pic->w;
            unsigned char *p2 = out->d + ((out->pitch - out->w) / 2 + x2 / scale) + out->pitch * (y2 / scale);
            for (;;) {
                if (e2 < e) {
                    memset(line, *p2++, e2 - line);
                    line = e2;
                } else {
                    memset(line, *p2, e - line);
                    break;
                }
                e2 += scale;
            }
        }
    } else {
        int y2 = y > height ? (y - height) / 2 : 0;
        int x2 = x > width ? (x - width) / 2 : 0;
        if (get_color_depth() <= 8) {
            static RGB_MAP rgb_table;
            generate_optimized_palette(pic, *screen_palette, NULL);
            create_rgb_table(&rgb_table, *screen_palette, NULL);
            if (!rgb_map) {lastblit.x = lastblit.y = 0; lastblit.w = SCREEN_W; lastblit.h = SCREEN_H;}
            rgb_map = &rgb_table;
        }
        for (y = 0; y < pic->h; y++, y2++) {
            create_tvpicture_line(out, vmode, x2, y2, pic->w, (unsigned char *)pic->line[y]);
        }
    }
    free_tv_pixels(out);
    return pic;
}

static int setvideomode(int hi) {
    static int bpp[5] = {32, 24, 16, 15, 8};
    int i;
#ifdef ALLEGRO_DOS
    int width = windim.w, height = windim.h;
    int driver = GFX_AUTODETECT;
#else
    int width = (arguments.fullscreen || windim.w > vidmode.w) ? vidmode.w : windim.w, height = (arguments.fullscreen || windim.h > vidmode.h) ? vidmode.h : windim.h;
    int driver = arguments.fullscreen ? GFX_AUTODETECT_FULLSCREEN : GFX_AUTODETECT_WINDOWED;
#endif

    for (i = (hi && !desktopdepth) ? 0 : 4; i < 5; i++) {
        if (!desktopdepth) set_color_depth(bpp[i]);
        if (!set_gfx_mode(driver, width, height, 0, 0)) {
            return 0;
        }
#ifndef ALLEGRO_DOS
        switch (driver) {
        case GFX_AUTODETECT_WINDOWED:
            if (!set_gfx_mode(GFX_AUTODETECT_FULLSCREEN, vidmode.w, vidmode.h, 0, 0)) {
                arguments.fullscreen = 1;
                return 0;
            }
            break;
        case GFX_AUTODETECT_FULLSCREEN:
            if (!set_gfx_mode(GFX_AUTODETECT_WINDOWED, windim.w > vidmode.w ? vidmode.w : windim.w, windim.h > vidmode.h ? vidmode.h : windim.h, 0, 0)) {
                arguments.fullscreen = 0;
                return 0;
            }
            break;
        }
#endif
    }
    set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
    return 1;
}

static void render_texture(view64_screen *rscreen, const rgb_color *palette) {
    static int texture_scale;
    static int fullscreen;
    static int olddepth;
    static unsigned int texture_video;
    static PALETTE screen_palette;
    static BITMAP *pic2 = NULL;
    int h, w, ph, pw, bx, by;
    int scale, scale2, picdepth;

    if (rscreen == NULL) {
        if (pic2) destroy_bitmap(pic2);
        return;
    }

    picdepth = colors_tvpicture(rscreen, arguments.vmode) > 256;
    if (!screen || olddim.w != windim.w || olddim.h != windim.h || (!desktopdepth && olddepth != picdepth) || fullscreen != arguments.fullscreen) {
        if (setvideomode(picdepth)) {
            return;
        }
        lastblit.w = 0;
        fullscreen = arguments.fullscreen;
        olddim = windim;
        olddepth = picdepth;
    }

    h = SCREEN_H; w = SCREEN_W;
    if (arguments.zoom) {
        scale = arguments.zoom;
    } else {
        scale = w / rscreen->w;
        scale2 = h / rscreen->h;
        if (scale2 < scale) scale = scale2;
        if (rscreen->w * scale > w || rscreen->h * scale > h) {
            scale--;
        }
        if (scale < 1) scale = 1;
    }

    pw = rscreen->w * scale;
    if (pw > SCREEN_W) pw = SCREEN_W;
    ph = rscreen->h * scale;
    if (ph > SCREEN_H) ph = SCREEN_H;

    if (texture_scale != scale || texture_video != arguments.vmode || pic2->w != pw || pic2->h != ph) {
        if (pic2) destroy_bitmap(pic2);
        pic2 = create_surface(rscreen, arguments.vmode, scale, palette, &screen_palette, SCREEN_W, SCREEN_H);
        if (pic2 == NULL) return;
        texture_scale = scale;
        texture_video = arguments.vmode;
        lastzoom = scale;
    }

    ph = pic2->h;
    pw = pic2->w;
    bx = (w - pw) / 2;
    by = (h - ph) / 2;

    set_palette(screen_palette);
    acquire_screen();
    if (lastblit.w) {
        int black = rgb_map ? makecol(0, 0, 0) : 0;
        if (by > lastblit.y) rectfill(screen, lastblit.x, lastblit.y, lastblit.x + lastblit.w - 1, by - 1, black);
        if (bx > lastblit.x) rectfill(screen, lastblit.x, by, bx - 1, ph + by - 1, black);
        if (lastblit.x + lastblit.w > bx + pw) rectfill(screen, bx + pw, by, lastblit.x + lastblit.w - 1, ph + by - 1, black);
        if (lastblit.y + lastblit.h > by + ph) rectfill(screen, lastblit.x, by + ph, lastblit.x + lastblit.w - 1, lastblit.y + lastblit.h - 1, black);
    }
    blit(pic2, screen, 0, 0, bx, by, pw, ph);
    lastblit.x = bx;
    lastblit.y = by;
    lastblit.w = pw;
    lastblit.h = ph;
    release_screen();
}

extern int main2(int, char *[]);

int main2(int argc, char *argv[])
{
    char cfgdir[MAX_PATH];
    unsigned char *mem;
    view64_screen *rscreen;
    FILE *f1;
    const char *filename = NULL;
    size_t size;
    view64_format format;
    rgb_color *palette;
    int done;

    config_reset(&arguments);
    get_user_config_file(cfgdir, sizeof(cfgdir), "view64", argv[0]);
    if (cfgdir[0]) {
        FILE *f1 = fopen2(cfgdir, "r");
        if (f1) {
            ini_parse_file(f1, ini_handler, &arguments);
            fclose(f1);
        }
    }
    configure(argc, argv, &arguments);
    if (arguments.zoom > 13) arguments.zoom = 13;
#ifdef ALLEGRO_DOS
    arguments.fullscreen = 1;
#endif
    filename = arguments.filename;

    if (!filename) {
        puts("Usage: view64 [OPTIONS...] PICTURE\n"
             "Try `view64 --help' or `view64 --usage' for more information.\n");
        free_config(&arguments);
        return EXIT_FAILURE;
    }

    if ((f1 = fopen2(filename, "rb")) == NULL) {
        fprintf(stderr, "Couldn't open file \"%s\"\n", filename);
        free_config(&arguments);
        return EXIT_FAILURE;
    }

    if (fseek(f1, 0, SEEK_END)) {
        size = 65536;
    } else {
        size = ftell(f1);
        if (size < 65536) size = 65536;
    }
    rewind(f1);
    mem = (unsigned char *)calloc(1, size);
    if (mem) size = fread(mem, 1, size, f1);
    fclose(f1);
    if (mem == NULL) {
        fprintf(stderr, "Out of memory.\n");
        free_config(&arguments);
        return EXIT_FAILURE;
    }

    format = view64_format_detect(filename, mem, size);
    rscreen = view64_decode(format, mem, size, 0);
    free(mem);

    if (format == VIEW64_UNKNOWN) {
        fprintf(stderr, "%s: Image format not recognized\n", filename);
        free_config(&arguments);
        return EXIT_FAILURE;
    }
    if (rscreen == NULL) {
        fprintf(stderr, "%s: Can't decode image\n", filename);
        free_config(&arguments);
        return EXIT_FAILURE;
    }

    if (allegro_init()) {
        fprintf(stderr, "Couldn't initialize allegro.\n");
        free_config(&arguments);
        return EXIT_FAILURE;
    }

    f1 = arguments.palette_filename && arguments.palette_filename[0] ? fopen2(arguments.palette_filename, "rb") : NULL;
    palette = load_palette(f1);
    if (f1) fclose(f1);

    if (arguments.save) {
        char *newname = (char *)malloc(strlen(arguments.filename) + 5);
        if (newname != NULL) {
            int zoom = arguments.zoom;
            static PALETTE surface_palette;
            BITMAP *surface;
            if (zoom < 1) {
                zoom = 1;
                while (rscreen->w * (zoom + 1) <= arguments.width && rscreen->h * (zoom + 1) <= arguments.height) zoom++;
            }
            surface = create_surface(rscreen, arguments.vmode, zoom, palette, &surface_palette, rscreen->w * zoom, rscreen->h * zoom);
            if (surface) {
                strcpy(newname, filename);
                strcat(newname, ".bmp");
                save_bmp(newname, surface, surface_palette);
                destroy_bitmap(surface);
            }
            free(newname);
        }
        view64_free_screen(rscreen);
        free(palette);
        free_config(&arguments);
        allegro_exit();
        return EXIT_SUCCESS;
    }

    if (get_desktop_resolution(&vidmode.w, &vidmode.h)) {
        vidmode.w = 640;
        vidmode.h = 480;
    }
    desktopdepth = desktop_color_depth();
    if (desktopdepth) set_color_depth(desktopdepth);
    set_color_conversion(COLORCONV_TOTAL | COLORCONV_DITHER_PAL);

    windim.w = arguments.width > 4160 ? 4160 : arguments.width;
    windim.h = arguments.height > 2600 ? 2600 : arguments.height;

    if (windim.w < 1) {
        if (windim.h < 1) {
            int scale = arguments.zoom;
#ifdef ALLEGRO_DOS
            if (!scale) scale = 1;
            windim.w = rscreen->w * scale;
            windim.h = rscreen->h * scale;
            if (windim.w <= 320 && windim.h <= 200) {
                windim.w = 320; windim.h = 200;
            } else if (windim.w <= 640 && windim.h <= 480) {
                windim.w = 640; windim.h = 480;
            } else if (windim.w <= 800 && windim.h <= 600) {
                windim.w = 800; windim.h = 600;
            } else {
                windim.w = 1024; windim.h = 768;
            }
#else
            windim.w = rscreen->w;
            windim.h = rscreen->h;
            if (!scale) {
                scale = 1;
                int sw = vidmode.w * 0.618, sh = vidmode.h * 0.618;
                while (windim.w * (scale + 1) < sw && windim.h * (scale + 1) < sh && scale < 13) {
                    scale++;
                }
            }
            windim.w *= scale;
            windim.h *= scale;
#endif
        } else {
            windim.w = rscreen->w * windim.h / rscreen->h;
        }
    } else if (windim.h < 1) {
        windim.h = rscreen->h * windim.w / rscreen->w;
    } else {
        /* all set */
    }

#ifndef ALLEGRO_DOS
    {
        const char *format_name = view64_format_name(format);
        char *window_title = (char *)malloc(strlen(format_name) + strlen("View64 " VERSION " - ") + 4 + strlen(filename));
        if (window_title == NULL) {
            fprintf(stderr, "Out of memory.\n");
            view64_free_screen(rscreen);
            free(palette);
            free_config(&arguments);
            allegro_exit();
            return EXIT_FAILURE;
        }
        strcpy(window_title, "View64 " VERSION " - ");
        strcat(window_title, format_name);
        strcat(window_title, " - ");
        strcat(window_title, filename);
        set_window_title(window_title);
        free(window_title);
    }
#endif

    install_keyboard(); 

    done = 0;
    do {
        int key;

        render_texture(rscreen, palette);
        key = readkey() & 0xff;
        if (key >= 'A' && key <= 'Z') key |= 0x20;

        switch (key) {
        case 13:
        case 27: done = 1; break;
        case 'p': arguments.vmode = (arguments.vmode & ~VIDEO_MASK) | VIDEO_PAL; break;
        case 'n': arguments.vmode = (arguments.vmode & ~VIDEO_MASK) | VIDEO_NTSC; break;
        case 'r': arguments.vmode = (arguments.vmode & ~VIDEO_MASK) | VIDEO_RGB; break;
        case 't': arguments.vmode ^= VIDEO_SHADOWMASK; break;
        case 'l': arguments.vmode ^= VIDEO_SCANLINES; break;
        case 'c': arguments.vmode ^= VIDEO_CHROMA_NOISE; break;
        case 'i': arguments.vmode ^= VIDEO_NOLACE; break;
        case 'm': arguments.vmode ^= VIDEO_MONOCHROME; break;
        case 'u': arguments.vmode ^= VIDEO_OLD_LUMA; break;
        case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8':
        case '9': arguments.zoom = key - '0'; break;
        case '=':
        case '+': if (lastzoom < 13) arguments.zoom = lastzoom + 1; break;
        case '-': if (lastzoom > 1) arguments.zoom = lastzoom - 1; break;
        case 's':
            {
                char *newname = (char *)malloc(strlen(arguments.filename) + 5);
                if (newname != NULL) {
                    static PALETTE surface_palette;
                    BITMAP *surface = create_surface(rscreen, arguments.vmode, lastzoom, palette, &surface_palette, rscreen->w * lastzoom, rscreen->h * lastzoom);
                    if (surface) {
                        strcpy(newname, filename);
                        strcat(newname, ".bmp");
                        save_bmp(newname, surface, surface_palette);
                        destroy_bitmap(surface);
                    }
                    free(newname);
                }
            }
            break;
#ifndef ALLEGRO_DOS
        case 'f': arguments.fullscreen ^= 1; break;
        case 'z':
            if (!arguments.fullscreen) {
                int h2, w2;
                if (arguments.zoom) {
                    w2 = rscreen->w * arguments.zoom;
                    h2 = rscreen->h * arguments.zoom;
                } else {
                    int scale = arguments.zoom;
                    w2 = rscreen->w;
                    h2 = rscreen->h;
                    if (!scale) {
                        scale = 1;
                        int sw = vidmode.w * 0.618, sh = vidmode.h * 0.618;
                        while (w2 * (scale + 1) < sw && h2 * (scale + 1) < sh && scale < 13) {
                            scale++;
                        }
                    }
                    w2 *= scale;
                    h2 *= scale;
                }
                windim.w = w2; windim.h = h2;
            }
#endif
        }
    } while (!done);

    render_texture(NULL, NULL);
    view64_free_screen(rscreen);
    free(palette);

    if (config_changed(&arguments)) {
        if (cfgdir[0]) {
            FILE *f1 = fopen2(cfgdir, "wt");
            if (f1) {
                ini_save_file(f1, &arguments);
                fclose(f1);
            }
        }
    }
    free_config(&arguments);
    set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
    allegro_exit();
    return EXIT_SUCCESS;
}
END_OF_MAIN()

